/*------------------------------------------------------------------------*/
/*  File: comp_feat.cpp                                                   */
/*  Feature computation for general MIP instances, including all types    */
/*  of instances CPLEX can read.                                          */
/*------------------------------------------------------------------------*/
/*  This code is a generalization and extension of feature computation    */
/*  for combinatorial auction instances written by Eugene Nudelman.       */
/*  The CPLEX interfacing is based on file examples/src/lpex2.c in        */
/*  CPLEX, Version 9.1                                                    */
/*  Frank Hutter, Aug 19, 2008                                            */
/*------------------------------------------------------------------------*/
/*  Differences in features compared to Nudelman's code are due to        */
/*  normalizations that were done in that code (but not reported in       */
/*  papers), and to at least one bug in that code: there is a problem in  */
/*  the bid graph degrees in that the code counts hundreds of times too   */
/*  high degrees, resulting in edge densities > 500 (should be            */
/*  upper-bounded by 1). Also, in one run valgrind found 7373 errors from */
/*  121 contexts, which suggests there may be some problem with memory.   */
/*------------------------------------------------------------------------*/
/*  I am being sloppy with freeing memory - this                          */
/*  is just fine when used as a standalone application, but if you want   */
/*  to use it in a bigger system you'll have to fix it to prevent leaks.  */
/*------------------------------------------------------------------------*/

/* Bring in the CPLEX function declarations and the C library 
   header file stdio.h with the following single include. */

#include <ilcplex/cplex.h>

/* Bring in the declarations for the string and character functions 
   and malloc, timer, and set. */

#include <ctype.h>
#include <cmath>
#include <stdlib.h>
#include <string.h>
#include "stopwatch.h"
#include <set>
#include <vector>
using namespace std;

// min and max macros
#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

/* Useful macro to swap 2 elements, using
 * the standard trick do { } while(0) to
 * make this one statement (otherwise it
 * could be wrong to use like 'else SWAP(A,B);').
 */
#define SWAP(A,B) \
do { int tmp = A; A = B; B = tmp; } while(0)

static void
   free_and_null (char **ptr),
   usage         (char *progname);


int size_of_intersection(int *a, int*b, int num_a, int num_b){
	/* Compute the size of the intersection of a[0],...,a[num_a-1], and b[0],...,b[num_b-1]
	   This assumes a and b are unique and sorted. */
	int i=0, j=0, jstart=0, count=0;
	while (i<num_a && j<num_b){
		if( a[i] > b[j] ){
			j++;
		} else if( a[i] < b[j] ){
			i++;
		} else {
			count++;
			i++;
			j++;
		}
	}
	return count;
}

void quick(int *a, int min, int max) {
  if (max - min > 1) {

    int i = min;
    int j = max;
    /* indices are positive */
    int pivot = a[(i+j) >> 1];
    do {
      while(a[i] < pivot) i++;
      while(a[j] > pivot) j--;
      if (i > j) break;
      SWAP(a[i], a[j]);
    } while(++i <= --j);

    /* Try to reduce bad behaviours. */
    while (min < j && a[j] == pivot) j--;
    if (min < j) quick(a, min, j);

    /* Try to reduce bad behaviours. */
    while (i < max && a[i] == pivot) i++;
    if (i < max) quick(a, i, max);

  } else if (a[min] > a[max])
    SWAP(a[min], a[max]);
}

/* Quick sort for sorting an int array a.
   Example: quick_sort(a, size);
// Equivalent: qsort(a, size, sizeof(int), compare_ints);
*/
void quick_sort(int *a, int size) {
  if (size > 1) {
    quick(a, 0, size-1);
  }
}

int compare_ints(const void * a, const void * b)
{
	const int *ia = (const int *) a;
	const int *ib = (const int *) b;
	
	return (*ia > *ib) - (*ia < *ib);
}


void printMatrixDouble(const double* X,int nRows, int nCols)
{
    int i,j;
    
    for(i = 0; i < nRows; i++) {
        printf("< ");
        for(j = 0; j < nCols; j++) {
            printf("%lf ",X[i+nRows*j]);}
        printf(">\n");}
}

void printMatrixInt(int* X,int nRows, int nCols)
{
    int i,j;
    
    for(i = 0; i < nRows; i++) {
        printf("< ");
        for(j = 0; j < nCols; j++) {
            printf("%d ",X[i+nRows*j]);}
        printf(">\n");}
}


double i_var(const int *array, const int N){
    int i;
    int sum_x, sum_x_squared;
	if(N<=1) return 0;
    sum_x = 0;
    sum_x_squared = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
        sum_x_squared += array[i]*array[i];
    }
/*
    printMatrixDouble(array, N, 1);
    printf("%lf, %lf, %d, %lf\n", sum_x, sum_x_squared, N, 1.0/(N-1) * (sum_x_squared - 1.0/N*sum_x*sum_x));
*/
    return fabs(1.0/(N-1) * (sum_x_squared - (1.0/N) * sum_x*sum_x));
}

double i_mean(const int *array, const int N){
    int i;
    int sum_x;
	if(N==0) return 0;
    sum_x = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
    }
    return sum_x/(N+0.0);
}

double d_var(const double *array, const int N){
    int i;
    double sum_x, sum_x_squared;
	if(N<=1) return 0;
    sum_x = 0;
    sum_x_squared = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
        sum_x_squared += array[i]*array[i];
    }

//    printMatrixDouble(array, N, 1);
/*    printf("%lf, %lf, %d, %lf\n", sum_x, sum_x_squared, N, 1.0/(N-1) * (sum_x_squared - 1.0/N*sum_x*sum_x));
*/
    return fabs(1.0/(N-1) * (sum_x_squared - (1.0/N) * sum_x*sum_x));
}

double d_mean(const double *array, const int N){
    int i;
    double sum_x;
	if(N==0) return 0;
    sum_x = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
    }
    return sum_x/(N+0.0);
}


int
main (int argc, char *argv[])
{
   /* Declare and allocate space for the variables and arrays. */
	set<int>::iterator it;
	vector< set<int> > vset;
	//set<int>::iterator it;
	Stopwatch sw;
	Stopwatch sw2;

   int      solnstat, solnmethod, solntype;
   double   *x     = NULL;
   int      *cstat = NULL;
   int      *rstat = NULL;
	char *xctype;

   CPXENVptr     env = NULL;
   CPXLPptr      lp = NULL;
   int           status = 0;
   int           i,j,k,j2,k2, for_up_bound, count;
   int           cur_numrows, cur_numcols;
   int           method;

   char          *basismsg;

	int *indices = new int[1];
	double *lu = new double[1];
	double *up = new double[1];
	double *lu_eps = new double[1];
	double *up_eps = new double[1];
	lu[0] = 0;
	up[0] = 1;

	char *Lchar = new char[1];
	char *Uchar = new char[1];
	char *Cchar = new char[1];
	Lchar[0] = 'L';
	Uchar[0] = 'U';
	Cchar[0] = 'C';

	int num_var_indices_to_make_cont;
    int *var_indices_to_make_cont;

	double sum = 0;
	double l2 = 0;
	double max_dist = -1;

	int nzcnt, nzcnt_returned, *rmatbeg, *rmatind, surplus, begin, end, num_nz_this_row;
	double *rmatval, *A_ij_normalized, *rhs_b, *a_normalized_varcoefs, a_ik, a_ij_magnitude_mean, a_ij_magnitude_std, a_ij_varcoef_avg, a_ij_varcoef_std, sum_abs;
	double *support_size, support_size_std, support_size_mean, perc_cont_var;
	int num_non_cont, probtype;

	int n_edges;
	double edge_density;
	int *vcg_var_degree, *vcg_constraint_degree, *vg_degree, *vg_degree_copy, **vg_neighbours;
	int vcg_var_deg_min, vcg_var_deg_max, vcg_con_deg_min, vcg_con_deg_max, vg_deg_max, vg_deg_min;
	double vcg_var_deg_mean, vcg_var_deg_std;
	double vcg_con_deg_mean, vcg_con_deg_std;
	double vg_deg_std, vg_deg_q0_25, vg_deg_med, vg_deg_q0_75;
	double *all_clust_coef, clust_coef, deviation, num_neighbour_edges;
	double *obj_fun_coef, prices_stddev, price_per_number_stddev, price_per_sqrt_number_stddev, *tmp_double_cur_numcols;
	double *up_bound, *low_bound;
	int num_q_constr, num_quad, num_qpnz, num_unbounded_discrete;
	double perc_unbounded_discrete;
	double lp_avg, lp_l2_avg, lp_linf, lp_objval;
	double time_parse, time_nonzero_enum, time_vg, time_clust_coef, time_relax, time_total;
	bool compute_nonzero_enum = true;
	bool compute_vg = true;
	bool compute_clust_coef = true;
	bool compute_relax = true;
	bool debug = false;

	sw.Start();
	sw2.Start();

	/* Check the command line arguments */
   if ( argc != 2 ) {
      usage (argv[0]);
      goto TERMINATE;
   }

   /* Initialize the CPLEX environment */

   env = CPXopenCPLEX (&status);

   /* If an error occurs, the status value indicates the reason for
      failure.  A call to CPXgeterrorstring will produce the text of
      the error message.  Note that CPXopenCPLEX produces no output,
      so the only way to see the cause of the error is to use
      CPXgeterrorstring.  For other CPLEX routines, the errors will
      be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON.  */

   if ( env == NULL ) {
      char  errmsg[1024];
      fprintf (stderr, "Could not open CPLEX environment.\n");
      CPXgeterrorstring (env, status, errmsg);
      fprintf (stderr, "%s", errmsg);
      goto TERMINATE;
   }

   /* Turn on output to the screen */

/*   status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
   if ( status ) {
      fprintf (stderr, 
               "Failure to turn on screen indicator, error %d.\n", status);
      goto TERMINATE;
   }*/

   /* Create the problem, using the filename as the problem name */

   lp = CPXcreateprob (env, &status, argv[1]);

   /* A returned pointer of NULL may mean that not enough memory
      was available or there was some other problem.  In the case of 
      failure, an error message will have been written to the error 
      channel from inside CPLEX.  In this example, the setting of
      the parameter CPX_PARAM_SCRIND causes the error message to
      appear on stdout.  Note that most CPLEX routines return
      an error code to indicate the reason for failure.   */

   if ( lp == NULL ) {
      fprintf (stderr, "Failed to create LP.\n");
      goto TERMINATE;
   }

   /* Now read the file, and copy the data into the created lp */

   status = CPXreadcopyprob (env, lp, argv[1], NULL);
   if ( status ) {
      fprintf (stderr, "Failed to read and copy the problem data.\n");
      goto TERMINATE;
   }
	time_parse = sw2.Lap();
	if (debug) fprintf(stderr, " parse time = %lfs\n", time_parse);

	/* Output: basic features */
	printf("n_constraints, n_variables, nzcnt");

	/* Output: QP features */
	printf(", probtype, num_q_constr, num_quad, num_qpnz");

	if (compute_nonzero_enum){
		/* Output: bipartite variable-constraint graph features; 4 features about degree of variables, 4 about constraints */
		printf(", vcg_var_deg_mean, vcg_var_deg_std, vcg_var_deg_min, vcg_var_deg_max, vcg_con_deg_mean, vcg_con_deg_std, vcg_con_deg_min, vcg_con_deg_max");

		/* Output: price-based features */
		printf(", prices_stddev, price_per_number_stddev, price_per_sqrt_number_stddev");

		/* Output: new features: percentage continuous variables, support size for discrete variables (mean, stddev), A matrix (4 feats) */
		printf(", perc_cont_var, support_size_mean, support_size_std, a_ij_magnitude_mean, a_ij_magnitude_std, a_ij_varcoef_avg, a_ij_varcoef_std, perc_unbounded_discrete");
	
		if (compute_vg){
			/* Output: variable graph (VG) features: edge density, degree statistics*/
			printf(", edge_density, vg_deg_std, vg_deg_max, vg_deg_min, vg_deg_q0_25, vg_deg_med, vg_deg_q0_75");

			if (compute_clust_coef){
				/* Output: more variable graph (VG) features: clustering coefficient and deviation*/
				printf(", clust_coef, deviation");
			}
		}
	}

	if( compute_relax ){
		/* Output: LP features */
	    printf(", lp_avg, lp_l2_avg, lp_linf, lp_objval\n");
	}

	
    /* The size of the problem should be obtained by asking CPLEX what
    the actual size is.  cur_numrows and cur_numcols store the 
    current number of rows and columns, respectively.  */
    cur_numcols = CPXgetnumcols (env, lp);
    cur_numrows = CPXgetnumrows (env, lp);
    nzcnt = CPXgetnumnz(env, lp);   /* this does *not* include the nonzero entries for quadratic constraints */

	/* Output: basic features */
	if (debug) fprintf(stderr, "Output: cur_numrows, cur_numcols, nzcnt\n");
	printf("%d, %d, %d", cur_numrows, cur_numcols, nzcnt);

	/* Compute features for quadratically-constrained problems */
	num_q_constr = CPXgetnumqconstrs (env, lp);
	num_quad = CPXgetnumquad (env, lp); // number of variables that have quadratic objective coefficients
	num_qpnz = CPXgetnumqpnz (env, lp);

	probtype = CPXgetprobtype (env, lp);
	//	printf("\nprobtype=%d\n",probtype);

	/* Output: QP features */
	if (debug) fprintf(stderr, "Output: probtype, num_q_constr, num_quad, num_qpnz\n");
	printf(", %d, %d, %d, %d", probtype, num_q_constr, num_quad, num_qpnz);

	if (compute_nonzero_enum){
		sw2.Start();
		fprintf(stderr, "Enumerating nonzero entries to compute VCG features and A_ij features - O(# nonzero entries) ...\n");

		/* Allocate memory for results of CPXgetrows. */
		rmatbeg = new int[cur_numrows];
		rmatind = new int[nzcnt];
		rmatval = new double[nzcnt];
		A_ij_normalized = new double[nzcnt];

		status = CPXgetrows(env, lp, &nzcnt_returned, rmatbeg, rmatind, rmatval, nzcnt, &surplus, 0, cur_numrows-1);
		if( nzcnt_returned != nzcnt ){
			printf("nzcnt = %d, nzcnt_returned = %d\n", nzcnt, nzcnt_returned);
			fprintf(stderr, "Error in computation of nonzero entries of the constraint matrix -- value returned by CPXgetrows different than the one from CPXgetnumnz\n" );
			goto TERMINATE;
		}

		if( surplus < 0 ){
			fprintf(stderr, "negative surplus %d, array too small\n", surplus);
			goto TERMINATE;
		} else {
	//		fprintf(stderr, "ok surplus %d\n", surplus);
		}
		
		/* Allocate memory for degree counters and lots of other arrays. */
		vcg_constraint_degree = new int[cur_numrows];
		rhs_b = new double[cur_numrows];
		a_normalized_varcoefs = new double[cur_numrows];

		vcg_var_degree = new int[cur_numcols];
		vg_degree = new int[cur_numcols];
		vg_degree_copy = new int[cur_numcols];
		vg_neighbours = new int*[cur_numcols];

		all_clust_coef = new double[cur_numcols];
		obj_fun_coef = new double[cur_numcols];
		tmp_double_cur_numcols = new double[cur_numcols];
		up_bound = new double[cur_numcols];
		low_bound = new double[cur_numcols];
		support_size = new double[cur_numcols];

		/* Initialize degree counters. */
		for (i=0; i<cur_numrows; i++) vcg_constraint_degree[i] = 0;
		for (k=0; k<cur_numcols; k++) vcg_var_degree[k] = 0;
		for (k=0; k<cur_numcols; k++) vg_degree[k] = 0;

		/* Get right hand side, the b array */
		status = CPXgetrhs (env, lp, rhs_b, 0, cur_numrows-1);
		fprintf(stderr, "#vars=%d, #constraints=%d, #nonzeros=%d\n", cur_numcols, cur_numrows, nzcnt);

		if ( status ) {
			fprintf (stderr, "Failed to get right hand side, the b array. Error %d\n", status);
			goto TERMINATE;
		}

		//printMatrixInt(vg_degree, 1, cur_numcols);
		vset = vector< set<int> > (cur_numcols);

		/* Enumerate nonzero entries, get variable and constraint degrees, and compute features of A matrix */
		count = 0;
		for(i=0; i<cur_numrows; i++){
			if (i==cur_numrows-1) {
				for_up_bound = nzcnt_returned; 	/* last row, special case for the upper limit of the loop */
			} else {
				for_up_bound = rmatbeg[i+1];
			}
			sum_abs = 0;
			for (j=rmatbeg[i]; j<for_up_bound; j++){
				k = rmatind[j];
				vcg_constraint_degree[i]++;
				vcg_var_degree[k]++;

				/* For computing features of the rows of the A matrix */
				a_ik = rmatval[j];
				tmp_double_cur_numcols[j-rmatbeg[i]] = fabs(a_ik); // absolute of nonzero entries in this row.
				sum_abs += fabs(a_ik);

				if (rhs_b[i] > 1e-6 || rhs_b[i] < -1e-6){
					/* A_ij_normalized holds constraint coefficients of normalized MIP with b = vector of ones (ignoring those rows with b=0 */
					A_ij_normalized[count++] = a_ik / rhs_b[i];
				}

				//printf(" row=%d, j=%d, k=%d, a_ik=%lf, rhs=%lf, begin=%d, end=%d,#vars in constraint=%d\n", i, j, k, a_ik, rhs_b[i], rmatbeg[i], for_up_bound , (for_up_bound - rmatbeg[i] - 1));
			}

			/* Compute variation coefficient of normalized absolute nonzero entries in this row */
			num_nz_this_row = for_up_bound-rmatbeg[i];
			for (j=0; j<num_nz_this_row; j++){
				tmp_double_cur_numcols[j] /= sum_abs;
			}
			if (num_nz_this_row == 0 ){
				a_normalized_varcoefs[i] = 0;
			} else {
				a_normalized_varcoefs[i] = sqrt(d_var(tmp_double_cur_numcols, num_nz_this_row)) / d_mean(tmp_double_cur_numcols, num_nz_this_row); 
			}
			//printf("num_nz_this_row=%d:\n",num_nz_this_row);
			//printMatrixDouble(tmp_double_cur_numcols, 1, num_nz_this_row);
			
			//printf("rmatbeg:\n");
			//printMatrixInt(rmatbeg, 1, cur_numrows);
			
		}

		//printMatrixDouble(a_normalized_varcoefs, 1, cur_numrows);
		a_ij_varcoef_avg = d_mean(a_normalized_varcoefs, cur_numrows);
		a_ij_varcoef_std = sqrt(d_var(a_normalized_varcoefs, cur_numrows));

		a_ij_magnitude_mean = d_mean(A_ij_normalized, count);
		a_ij_magnitude_std = sqrt(d_var(A_ij_normalized, count));


		/* Bipartite variable-constraint graph (VCG); 4 features about degree of variables, 4 about constraints */
		
		/* 1. Degree of variable nodes */
		vcg_var_deg_mean = i_mean(vcg_var_degree, cur_numcols);
		vcg_var_deg_std = sqrt(i_var(vcg_var_degree, cur_numcols));
		vcg_var_deg_min = cur_numrows+1;
		vcg_var_deg_max = -1;
		for(i=0; i<cur_numcols; i++){ 
			if( vcg_var_degree[i] < vcg_var_deg_min ) vcg_var_deg_min = vcg_var_degree[i];
			if( vcg_var_degree[i] > vcg_var_deg_max ) vcg_var_deg_max = vcg_var_degree[i];
			if (vcg_var_degree[i] == 0){
				fprintf(stderr, "Warning: variable %d is not mentioned in any linear constraints ... \n", i);
			}
		}

		/* Variable-constraint graph features, degree of variable nodes */
		/* FH: different from the original WDP feature code, I don't normalize these by NrowConstraints -- for std, this normalization was also omitted (forgotten?) in the original code. */
		if (debug) fprintf(stderr, "Output: vcg_var_deg_mean, vcg_var_deg_std, vcg_var_deg_min, vcg_var_deg_max\n");
		printf(", %lf, %lf, %d, %d", vcg_var_deg_mean, vcg_var_deg_std, vcg_var_deg_min, vcg_var_deg_max);

		/* 2. Degree of constraint nodes */
		vcg_con_deg_mean = i_mean(vcg_constraint_degree, cur_numrows);
		vcg_con_deg_std = sqrt(i_var(vcg_constraint_degree, cur_numrows));
		vcg_con_deg_min = cur_numcols+1;
		vcg_con_deg_max = -1;
		for(j=0; j<cur_numrows; j++){ 
			if( vcg_constraint_degree[j] < vcg_con_deg_min ) vcg_con_deg_min = vcg_constraint_degree[j];
			if( vcg_constraint_degree[j] > vcg_con_deg_max ) vcg_con_deg_max = vcg_constraint_degree[j];
			if (vcg_constraint_degree[j] == 0){
				fprintf(stderr, "Warning: constraint %d does not include any variables ... \n", j);
			}
		}

		/* Variable-constraint graph features, degree of constraint nodes */
		/* FH: different from the original WDP feature code, I don't normalize these by McolumnVars. */
		/* Output: bipartite variable-constraint graph features; 4 features about degree of variables, 4 about constraints */
		if (debug) fprintf(stderr, "Output: vcg_con_deg_mean, vcg_con_deg_std, vcg_con_deg_min, vcg_con_deg_max\n");
		printf(", %lf, %lf, %d, %d", vcg_con_deg_mean, vcg_con_deg_std, vcg_con_deg_min, vcg_con_deg_max);


		/* Compute features using objective function coefficients */
		status = CPXgetobj (env, lp, obj_fun_coef, 0, cur_numcols-1);
		prices_stddev = sqrt(d_var(obj_fun_coef, cur_numcols));

		for(k=0; k<cur_numcols; k++){
			if (vcg_var_degree[k] == 0){
				tmp_double_cur_numcols[k] = obj_fun_coef[k];
			} else {
				tmp_double_cur_numcols[k] = obj_fun_coef[k] / vcg_var_degree[k];
			}
		}
		price_per_number_stddev = sqrt(d_var(tmp_double_cur_numcols, cur_numcols));

		for(k=0; k<cur_numcols; k++){
			if (vcg_var_degree[k] == 0){
				tmp_double_cur_numcols[k] = obj_fun_coef[k];
			} else {
				tmp_double_cur_numcols[k] = obj_fun_coef[k] / sqrt(vcg_var_degree[k]);
			}
		}
		price_per_sqrt_number_stddev = sqrt(d_var(tmp_double_cur_numcols, cur_numcols));

		/* Output: price-based features */
		if (debug) fprintf(stderr, "Output: prices_stddev, price_per_number_stddev, price_per_sqrt_number_stddev\n");
		printf(", %lf, %lf, %lf", prices_stddev, price_per_number_stddev, price_per_sqrt_number_stddev);


		/* Compute new features */
		status = CPXgetub(env, lp, up_bound, 0, cur_numcols-1);
		status = CPXgetlb(env, lp, low_bound, 0, cur_numcols-1);

		if (probtype == CPXPROB_LP || probtype == CPXPROB_QP || probtype == CPXPROB_QCP){
			/* These problem types have only continuous variables */
			support_size_mean = 0;
			support_size_std = 0;
			perc_unbounded_discrete = 0;
			perc_cont_var = 0;
		} else {
			/* Remember which variables are integer valued -- only compute integer slack for those later */
   			xctype = new char[cur_numcols];
			status = CPXgetctype (env, lp, xctype, 0, cur_numcols-1);

			if ( status ) {
				fprintf (stderr, "Failed to get the type of variables, error %d.\n", status);
				goto TERMINATE;
			}

			/* Percentage continuous or semi-continuous variables, and support sizes for discrete variables. */
			count = 0;
			num_non_cont = 0;
			num_unbounded_discrete = 0;
			for(i=0; i<cur_numcols; i++){
				if (xctype[i] == 'C' || xctype[i] == 'S'){
					count++;
				} else {
					if (xctype[i] == 'B'){
						support_size[num_non_cont++] = 1+1;
					} else {
						if (up_bound[i] == CPX_INFBOUND){
							num_unbounded_discrete++;
						} else if (low_bound[i] == -CPX_INFBOUND){
							num_unbounded_discrete++;
						} else {
							support_size[num_non_cont++] = up_bound[i]-low_bound[i]+1;
						}
					}
				}
			}
			perc_unbounded_discrete = (num_unbounded_discrete+0.0)/(num_non_cont+num_unbounded_discrete);
			perc_cont_var = (count + 0.0) / cur_numcols;
		//	printf("up[0]=%lf, low[0]=%lf\n", up_bound[0], low_bound[0]);
		//	printMatrixDouble(support_size, 1, num_non_cont);
			support_size_mean = d_mean(support_size, num_non_cont);
			support_size_std = sqrt(d_var(support_size, num_non_cont));
		}

		/* Output: new features: percentage continuous variables, support size for discrete variables (mean, stddev), A matrix (4 feats) */
		if (debug) fprintf(stderr, "Output: perc_cont_var, support_size_mean, support_size_std, a_ij_magnitude_mean, a_ij_magnitude_std, a_ij_varcoef_avg, a_ij_varcoef_std, perc_unbounded_discrete\n");
		printf(", %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf", perc_cont_var, support_size_mean, support_size_std, a_ij_magnitude_mean, a_ij_magnitude_std, a_ij_varcoef_avg, a_ij_varcoef_std, perc_unbounded_discrete);

		time_nonzero_enum = sw2.Lap();
		if (debug) fprintf(stderr, " enumeration of nonzero elements took %lfs\n", time_nonzero_enum);

		if (compute_vg){
			sw2.Start();
			fprintf(stderr, "Computing VG features - O(#num constraints * (#num vars in constraint)^2) ...\n");
			for(i=0; i<cur_numrows; i++){
				if (i==cur_numrows-1) {
					for_up_bound = nzcnt_returned-1; 	/* last row, special case for the upper limit of the loop */
				} else {
					for_up_bound = rmatbeg[i+1];
				}
				sum_abs = 0;
				for (j=rmatbeg[i]; j<for_up_bound; j++){
					k = rmatind[j];
					/* Collect neighbours */
					for (j2=j+1; j2<for_up_bound; j2++){
						k2 = rmatind[j2];
						vset[ k ].insert( k2 );
						vset[ k2 ].insert( k );
					}
				}
			}

			for(k=0; k<cur_numcols; k++){
				vg_degree[k] = vset[ k ].size();
				vg_neighbours[k] = new int[vg_degree[k]];
				count = 0;
				for (it=vset[ k ].begin(); it!=vset[ k ].end(); it++){
					vg_neighbours[k][count++] = *it;
				}
			}
			// printMatrixInt(vg_degree, 1, cur_numcols);

			vg_deg_min = cur_numrows+1;
			vg_deg_max = -1;
			n_edges = 0;
			for(k=0; k<cur_numcols; k++){
				if( vg_degree[k] < vg_deg_min ){
					vg_deg_min = vg_degree[k];
				}
				if( vg_degree[k] > vg_deg_max ){
					vg_deg_max = vg_degree[k];
				}
				n_edges += vg_degree[k];
			}
			n_edges /= 2;
			edge_density = n_edges / (double)( (cur_numcols*(cur_numcols-1)) / 2 );
		//	printf("n_edges=%d, full=%lf\n", n_edges, (double)( (cur_numcols*(cur_numcols-1)) / 2 ));
		//	vg_deg_mean = i_mean(vg_degree, cur_numcols);
			vg_deg_std = sqrt(i_var(vg_degree, cur_numcols));

			for (k=0; k<cur_numcols; k++){
				vg_degree_copy[k] = vg_degree[k];
			}

			/* Sorting vg_degree_copy, so it's not matching with the columns anymore! */
			quick_sort(vg_degree_copy, cur_numcols); 
			vg_deg_q0_25 = vg_degree_copy [cur_numcols / 4];
			vg_deg_med = vg_degree_copy [2*cur_numcols / 4];
			vg_deg_q0_75 = vg_degree_copy [3*cur_numcols / 4];

			/* Output: variable graph (VG) features: edge density, degree statistics */
			if (debug) fprintf(stderr, "Output: edge_density, vg_deg_std, vg_deg_max, vg_deg_min, vg_deg_q0_25, vg_deg_med, vg_deg_q0_75\n");
			printf(", %lf, %lf, %d, %d, %lf, %lf, %lf", edge_density, vg_deg_std, vg_deg_max, vg_deg_min, vg_deg_q0_25, vg_deg_med, vg_deg_q0_75);

			time_vg = sw2.Lap();
			if (debug) fprintf(stderr, " VG features took %lfs\n", time_vg);

			if (compute_clust_coef){
				sw2.Start();
				fprintf(stderr, "Computing clustering coefficient - O(#variables^3) ...\n");
				/* Compute clustering coefficient and deviation */
				for (k=0; k<cur_numcols; k++){
					num_neighbour_edges = 0;
					if( vg_degree[k] <= 1 ){
						all_clust_coef[k] = 0;
					} else {
						for (i=0; i<vg_degree[k]; i++){
							/* For each variable k and each of its neighbours k2': intersect N(k) and N(k2') (which are already sorted), divide by 2 */
							k2 = vg_neighbours[k][i];
							num_neighbour_edges += size_of_intersection(vg_neighbours[k], vg_neighbours[k2], vg_degree[k], vg_degree[k2]);
						}
						num_neighbour_edges /= 2;
						all_clust_coef[k] = num_neighbour_edges/double(vg_degree[k]*(vg_degree[k]-1)/2);
					}
				}
			//	printMatrixDouble(all_clust_coef, 1, cur_numcols);

				clust_coef = d_mean(all_clust_coef, cur_numcols);
				deviation = sqrt(d_var(all_clust_coef, cur_numcols));

				/* Output: more variable graph (VG) features: clustering coefficient and deviation*/
				if (debug) fprintf(stderr, "Output: clust_coef, deviation\n");
				printf(", %lf, %lf", clust_coef, deviation);

				time_clust_coef = sw2.Lap();
				if (debug) fprintf(stderr, " clustering coefficient took %lfs\n", time_clust_coef);
			}
		}
	}

	if (compute_relax){
		sw2.Start();
		/* Rest of the file: LP features. */
		if (probtype == CPXPROB_LP || probtype == CPXPROB_QP || probtype == CPXPROB_QCP){
			/* These problem types have only continuous variables, so zero LP slack; using lp_objval=0 here */
			lp_avg = 0;
			lp_l2_avg = 0;
			lp_linf = 0;
			lp_objval = 0;
		} else {
			var_indices_to_make_cont = new int[cur_numcols];
			num_var_indices_to_make_cont = 0;

			for(i=0; i<cur_numcols; i++){
				//printf("xctype[i] = %c, low_bound[i]=%lf\n", xctype[i], low_bound[i]);
				if (xctype[i] != 'C' && xctype[i] != 'S'){ // if not (continuous or semi-continuous)
					if (xctype[i] == 'B'){
						indices[0] = i;
						/* For previously binary variables, set bounds to [0,1] */
						status = CPXchgbds (env, lp, 1, indices, Lchar, lu);
						status = CPXchgbds (env, lp, 1, indices, Uchar, up);
						low_bound[i] = lu[0];
						up_bound[i] = up[0];
					}
					var_indices_to_make_cont[num_var_indices_to_make_cont++] = i;

					/* Set lower bound eps lower, upper bound eps higher to avoid numerical problems. */
					indices[0] = i;
					lu_eps[0] = low_bound[i]; //-CPX_INFBOUND; //
					up_eps[0] = up_bound[i]; //CPX_INFBOUND;//
					status = CPXchgbds (env, lp, 1, indices, Lchar, lu_eps);
					status = CPXchgbds (env, lp, 1, indices, Uchar, up_eps);
				}
			}


			/* Relax problem: change all variables to continuous, then solve as LP (or QP/QCP) */

			/* Set LP strategy, optimize the LP and obtain solution. */
			if (probtype == CPXPROB_MILP || probtype == CPXPROB_FIXEDMILP){
				status = CPXchgprobtype (env, lp, CPXPROB_LP);
				status = CPXsetintparam (env, CPX_PARAM_LPMETHOD, CPX_ALG_AUTOMATIC);
				if ( status ) {
					fprintf (stderr, 
							"Failed to set LP optimization method, error %d.\n", status);
					goto TERMINATE;
				}
				status = CPXlpopt (env, lp);
				if ( status ) {
					fprintf (stderr, "Failed to optimize LP.\n");
					goto TERMINATE;
				}

			} else if (probtype == CPXPROB_MIQP || probtype == CPXPROB_FIXEDMIQP){
				status = CPXchgprobtype (env, lp, CPXPROB_QP);
				status = CPXsetintparam (env, CPX_PARAM_QPMETHOD, CPX_ALG_AUTOMATIC);
				if ( status ) {
					fprintf (stderr, 
							"Failed to set QP optimization method, error %d.\n", status);
					goto TERMINATE;
				}
				status = CPXqpopt (env, lp);
				if ( status ) {
					fprintf (stderr, "Failed to optimize QP.\n");
					goto TERMINATE;
				}

			} else {
				status = CPXchgprobtype (env, lp, CPXPROB_QCP);
				status = CPXsetintparam (env, CPX_PARAM_LPMETHOD, CPX_ALG_AUTOMATIC);
				if ( status ) {
					fprintf (stderr, 
							"Failed to set QCP optimization method, error %d.\n", status);
					goto TERMINATE;
				}
				status = CPXbaropt (env, lp);
				if ( status ) {
					fprintf (stderr, "Failed to optimize QCP.\n");
					goto TERMINATE;
				}

			}

		    solnstat = CPXgetstat (env, lp);
			if      ( solnstat == CPX_STAT_UNBOUNDED ) {
				fprintf (stderr, "Relaxed model is unbounded\n");
				goto TERMINATE;
			}
			else if ( solnstat == CPX_STAT_INFEASIBLE ) {
				fprintf (stderr, "Relaxed model could not be shown to be feasible\n");
				goto TERMINATE;
			}
			else if ( solnstat == CPX_STAT_INForUNBD ) {
				printf ("Relaxed model is infeasible or unbounded\n");
				goto TERMINATE;
			} 
			else {
				status = CPXsolninfo (env, lp, &solnmethod, &solntype, NULL, NULL);
				if ( status ) {
					fprintf (stderr, "Failed to obtain LP solution info.\n");
					goto TERMINATE;
				}
				//printf ("Solution status %d, solution method %d\n", solnstat, solnmethod);
			}

			if ( solntype == CPX_NO_SOLN ) {
				fprintf (stderr, "Relaxed solution not available.\n");
				//goto TERMINATE;
			}
			    
			status = CPXgetobjval (env, lp, &lp_objval);
			if ( status ) {
				fprintf (stderr, "Failed to obtain relaxed objective value.\n"); 
				// FH: if this error occurs for an instance, one could code "failed objective" with a special value
				goto TERMINATE;
			}
			//printf ("Objective value %lf\n", lp_objval);

			/* Retrieve solution vector */
			x = (double *) malloc (cur_numcols*sizeof(double));
			if ( x == NULL ) {
				fprintf (stderr, "No memory for solution.\n");
				goto TERMINATE;
			}

			status = CPXgetx (env, lp, x, 0, cur_numcols-1);
			if ( status ) {
				fprintf (stderr, "Failed to obtain primal solution.\n");
				goto TERMINATE;
			}

			for(j=0; j<num_var_indices_to_make_cont; j++) {
				i = var_indices_to_make_cont[j];
				double dist = min(fabs(x[i]), fabs(x[i] - 1));
				sum += dist;
				l2 += dist*dist;
				max_dist = max(max_dist, dist);
			}

			lp_avg = sum / double(num_var_indices_to_make_cont);
			lp_l2_avg = sqrt(l2/double(num_var_indices_to_make_cont));
			lp_linf = max_dist;
		}
		/* Output: LP features */
		if (debug) fprintf(stderr, "Output: lp_avg, lp_l2_avg, lp_linf, lp_objval\n");
		printf(", %lf, %lf, %lf, %lf\n", lp_avg, lp_l2_avg, lp_linf, lp_objval);

		time_relax = sw2.Lap();
		if (debug) fprintf(stderr, " Solving relaxed problem took %lfs\n", time_relax);
	}

   /* Free up the basis and solution */

   free_and_null ((char **) &cstat);
   free_and_null ((char **) &rstat);
   free_and_null ((char **) &x);

   /* Free up the problem, if necessary */

   if ( lp != NULL ) {
      status = CPXfreeprob (env, &lp);
      if ( status ) {
         fprintf (stderr, "CPXfreeprob failed, error code %d.\n", status);
      }
   }

   /* Free up the CPLEX environment, if necessary */

   if ( env != NULL ) {
      status = CPXcloseCPLEX (&env);

      /* Note that CPXcloseCPLEX produces no output,
         so the only way to see the cause of the error is to use
         CPXgeterrorstring.  For other CPLEX routines, the errors will
         be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON. */

      if ( status ) {
         char  errmsg[1024];
         fprintf (stderr, "Could not close CPLEX environment.\n");
         CPXgeterrorstring (env, status, errmsg);
         fprintf (stderr, "%s", errmsg);
      }
   }

	time_total = sw.Lap();
	printf("time_parse, time_enum_nz, time_vg, time_clust_coef, time_relax, time_total\n");
	printf("%lf, %lf, %lf, %lf, %lf, %lf\n", time_parse, time_nonzero_enum, time_vg, time_clust_coef, time_relax, time_total);
	printf("Total feature computation time: %lfs", time_total);

TERMINATE:
	printf("\n");
   return (status);

}  /* END main */




/* This simple routine frees up the pointer *ptr, and sets *ptr to NULL */

static void
free_and_null (char **ptr)
{
   if ( *ptr != NULL ) {
      free (*ptr);
      *ptr = NULL;
   }
} /* END free_and_null */


static void
usage (char *progname)
{
   fprintf (stderr,"Usage: %s filename\n", progname);
   fprintf (stderr,"   where filename is a MIP instance file in .MPS, .LP, or .SAV format.\n");
   fprintf (stderr," Exiting...\n");
} /* END usage */
