/*------------------------------------------------------------------------*/
/*  File: comp_feat.cpp                                                   */
/*  Feature computation for general MIP instances, including all types    */
/*  of instances CPLEX can read.                                          */
/*------------------------------------------------------------------------*/
/*  This code started as a generalization of feature computation code     */
/*  for combinatorial auction instances written by Eugene Nudelman.       */
/*  Since then, we added lots of new features.                            */
/*  The CPLEX interfacing is based on file examples/src/lpex2.c in        */
/*  CPLEX, Version 9.1                                                    */
/*  Frank Hutter, Aug 19, 2008                                            */
/*------------------------------------------------------------------------*/
/*  Differences in features compared to Nudelman's code are due to        */
/*  normalizations that were done in that code (but not reported in       */
/*  papers), and to at least one bug in that code: there is a problem in  */
/*  the bid graph degrees in that the code counts hundreds of times too   */
/*  high degrees, resulting in edge densities > 500 (should be            */
/*  upper-bounded by 1). Also, in one run valgrind found 7373 errors from */
/*  121 contexts, which suggests there may be some problem with memory.   */
/*------------------------------------------------------------------------*/
/*  I am being sloppy with freeing memory - this                          */
/*  is just fine when used as a standalone application, but if you want   */
/*  to use it in a bigger system you'll have to fix it to prevent leaks.  */
/*------------------------------------------------------------------------*/



/* Bring in the declarations for the string and character functions 
   and malloc, timer, and set. */

#include "comp_feat.h"
#include <boost/lexical_cast.hpp>
using namespace std;


// min and max macros
#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

/* Useful macro to swap 2 elements, using
 * the standard trick do { } while(0) to
 * make this one statement (otherwise it
 * could be wrong to use like 'else SWAP(A,B);').
 */
#define SWAP(A,B) \
do { int tmp = A; A = B; B = tmp; } while(0)

void writefeat(string name, double value){
   featname.push_back(name);
   featvalue.push_back(value);
}

void printfeat(){
     vector<string>::iterator i;
     string tmpname;
     vector<double>::iterator j;
     double tmpvalue;
     for (i = featname.begin(); i != featname.end()-1; i++) {
       tmpname=*i;
       cout << tmpname <<", ";
     } 
       tmpname=featname.back();
       cout << tmpname<< endl;
     for (j = featvalue.begin(); j != featvalue.end()-1; j++) {
       tmpvalue=*j;
       cout << tmpvalue <<", ";
     } 
       tmpvalue=featvalue.back();
       cout << tmpvalue<< endl;
}

void terminate(int status){
	printfeat();
	cout << "\n";
	exit(status);
}

string convertInt(int number){
	return boost::lexical_cast<string>( number );
}

static void
   free_and_null (char **ptr),
   usage         (char *progname);


int size_of_intersection(int *a, int*b, int num_a, int num_b){
	/* Compute the size of the intersection of a[0],...,a[num_a-1], and b[0],...,b[num_b-1]
	   This assumes a and b are unique and sorted. */
	int i=0, j=0, jstart=0, count=0;
	while (i<num_a && j<num_b){
		if( a[i] > b[j] ){
			j++;
		} else if( a[i] < b[j] ){
			i++;
		} else {
			count++;
			i++;
			j++;
		}
	}
	return count;
}

void quick(int *a, int min, int max) {
	if (max - min > 1) {
		int i = min;
		int j = max;
		/* indices are positive */
		int pivot = a[(i+j) >> 1];
		do {
			while(a[i] < pivot) i++;
			while(a[j] > pivot) j--;
			if (i > j) break;
			SWAP(a[i], a[j]);
		} while(++i <= --j);

		/* Try to reduce bad behaviours. */
		while (min < j && a[j] == pivot) j--;
			if (min < j) quick(a, min, j);

		/* Try to reduce bad behaviours. */
		while (i < max && a[i] == pivot) i++;
			if (i < max) quick(a, i, max);

	} else if (a[min] > a[max])
		SWAP(a[min], a[max]);
}

/* Quick sort for sorting an int array a.
   Example: quick_sort(a, size);
// Equivalent: qsort(a, size, sizeof(int), compare_ints);
*/
void quick_sort(int *a, int size) {
	if (size > 1) {
		quick(a, 0, size-1);
	}
}

int compare_ints(const void * a, const void * b){
	const int *ia = (const int *) a;
	const int *ib = (const int *) b;
	return (*ia > *ib) - (*ia < *ib);
}

void printMatrixDouble(const double* X,int nRows, int nCols){
    int i,j;
    for(i = 0; i < nRows; i++) {
        printf("< ");
        for(j = 0; j < nCols; j++) {
            printf("%lf ",X[i+nRows*j]);
		}
        printf(">\n");
	}
}

void printMatrixInt(int* X,int nRows, int nCols){
    int i,j;
    for(i = 0; i < nRows; i++) {
        printf("< ");
        for(j = 0; j < nCols; j++) {
            printf("%d ",X[i+nRows*j]);
		}
        printf(">\n");
	}
}

double d_var(const double *array, const int N){
    int i;
    double sum_x, sum_x_squared;
	if(N<=1) return 0.0;
    sum_x = 0;
    sum_x_squared = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
        sum_x_squared += array[i]*array[i];
    }
    return fabs(1.0/(N-1) * (sum_x_squared - (1.0/N) * sum_x*sum_x));
}

double d_mean(const double *array, const int N){
    int i;
    double sum_x;
	if(N==0) return 0;
    sum_x = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
    }
    return sum_x/(N+0.0);
}

double d_var_coef(const double *array, const int N){
    int i;
	if(N<=1) return 0;
    return sqrt(d_var(array, N)) / d_mean(array, N);
}

void compute_array_stats(const double *array, const int N, const string array_name, double* results, string* result_names){
//	printf("N in compute_array_stats: %d\n",N);
	result_names[0] = array_name + "__mean";
	results[0] = d_mean(array, N);
	
//	result_names[1] = array_name + "__varcoef";
//	results[1] = d_var_coef(array, N);

//	printf("In compute_array_stats, N=%d\n",N);
//    printMatrixDouble(array, N, 1);
//	printf("%lf \n", array[0]);


	result_names[1] = array_name + "__std";

	results[1] = sqrt(d_var(array, N));

//	cout << result_names[0] << " " << results[0] << "\n";
}

int initialize(char* filename, int debug, CPXENVptr *env, CPXLPptr *lp){
	int status;
	/* Initialize the CPLEX environment */
	*env = CPXopenCPLEX (&status);
	/* If an error occurs, the status value indicates the reason for
	  failure.  A call to CPXgeterrorstring will produce the text of
	  the error message.  Note that CPXopenCPLEX produces no output,
	  so the only way to see the cause of the error is to use
	  CPXgeterrorstring.  For other CPLEX routines, the errors will
	  be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON.  */
	if ( *env == NULL ) {
		char  errmsg[1024];
		fprintf (stderr, "Could not open CPLEX environment.\n");
		CPXgeterrorstring (*env, status, errmsg);
		fprintf (stderr, "%s", errmsg);
		terminate(status);
	}

	/* Set node log interval. */
	status = CPXsetintparam (*env, CPX_PARAM_MIPINTERVAL, 1);
	if ( status ) {
		fprintf (stderr, 
                 "Failure to set CPX_PARAM_MIPINTERVAL, error %d.\n", status);
		terminate(status);
	}

	/* Set level pf detail for node log. */
	status = CPXsetintparam (*env, CPX_PARAM_MIPDISPLAY, 2);
	if ( status ) {
		fprintf (stderr, 
                 "Failure to set CPX_PARAM_MIPDISPLAY, error %d.\n", status);
		terminate(status);
	}

	/* Turn on output to the screen */
	/*   status = CPXsetintparam (*env, CPX_PARAM_SCRIND, CPX_ON);
	if ( status ) {
	  fprintf (stderr, 
			   "Failure to turn on screen indicator, error %d.\n", status);
	  terminate(status);
	}*/

	/* Create the problem, using the filename as the problem name */
	*lp = CPXcreateprob (*env, &status, filename);
	/* A returned pointer of NULL may mean that not enough memory
	  was available or there was some other problem.  In the case of 
	  failure, an error message will have been written to the error 
	  channel from inside CPLEX.  In this example, the setting of
	  the parameter CPX_PARAM_SCRIND causes the error message to
	  appear on stdout.  Note that most CPLEX routines return
	  an error code to indicate the reason for failure.   */
	if ( *lp == NULL ) {
	  fprintf (stderr, "Failed to create LP.\n");
	  terminate(status);
	}

    /* Now read the file, and copy the data into the created lp */
	status = CPXreadcopyprob (*env, *lp, filename, NULL);
	if ( status ) {
		fprintf (stderr, "Failed to read and copy the problem data.\n");
		terminate(status);
	}
	double time_parse = sw.Lap();
	if (debug) fprintf(stderr, " parse time = %lfs\n", time_parse);
}

void getConstraintTypes(CPXENVptr env, CPXLPptr lp, int cur_numrows, vector< set<int> > *constraint_sets){
	char* sense = new char[cur_numrows];

	/* Get sense of constraints. */
	int status = CPXgetsense (env, lp, sense, 0, cur_numrows-1);
	if ( status ) {
		fprintf (stderr, "Failed to get constraint sense. Error %d\n", status);
		terminate(status);
	}

	(*constraint_sets) = vector< set<int> > (4); /* 'L', 'E', 'G', 'R' */
	for(int i=0; i<cur_numrows; i++){
		switch ( sense[i] ) {
			case 'L':
				(*constraint_sets)[0].insert(i);
				break;
			case 'E':
				(*constraint_sets)[1].insert(i);
				break;
			case 'G':
				(*constraint_sets)[2].insert(i);
				break;
			case 'R':
				(*constraint_sets)[3].insert(i);
				break;			
		}
	}
}


void getVariableTypes(CPXENVptr env, CPXLPptr lp, int probtype, int cur_numcols, vector< set<int> > *variable_sets){
	int i;
	int num_unbounded_discrete, num_non_cont;
	double *up_bound, *low_bound;

	/* Get upper and lower bounds. */
	up_bound = new double[cur_numcols];
	low_bound = new double[cur_numcols];
	int status = CPXgetub(env, lp, up_bound, 0, cur_numcols-1);
	status = CPXgetlb(env, lp, low_bound, 0, cur_numcols-1);
	
	/* Get variable types in xctype. */
	char *xctype = new char[cur_numcols];
	if (probtype == 0){
		/* LPs don't have ctype information, so manually fill it in. */
		for(i=0; i<cur_numcols; i++){
			xctype[i] = 'C';
		}
	} else {
		status = CPXgetctype (env, lp, xctype, 0, cur_numcols-1);
		if ( status ) {
			fprintf (stderr, "Failed to get the type of variables, error %d.\n", status);
			terminate(status);
		}
	}
	
	/* Collect integer and continuous variables into sets */
	*variable_sets = vector< set<int> > (3); /* continuous, integer&binary, all */
	int num_binary = 0;
	int num_integer = 0;
	int num_continuous = 0;
	int num_semicont = 0;
	int num_semiint = 0;
	
	for(i=0; i<cur_numcols; i++){
		switch ( xctype[i] ) {
			case 'B':
				num_binary++;
				break;
			case 'I':
				num_integer++;
				break;
			case 'C':
				num_continuous++;
				break;
			case 'S':
				num_semicont++;
				break;
			case 'N':
				num_semiint++;
				break;			
		}
		if (xctype[i] == 'C'){ /* not counting semi-continuous variables here! xctype[i] == 'S' */
			(*variable_sets)[1].insert(i);
		} else {
			(*variable_sets)[0].insert(i);
		}
		(*variable_sets)[2].insert(i);
	}
	
	/* Output: basic features */
	writefeat("num_b_variables", num_binary);
	writefeat("num_i_variables", num_integer);
	writefeat("num_c_variables", num_continuous);
	writefeat("num_s_variables", num_semicont);
	writefeat("num_n_variables", num_semiint);
	writefeat("ratio_b_variables", num_binary / (cur_numcols+0.0));
	writefeat("ratio_i_variables", num_integer / (cur_numcols+0.0));
	writefeat("ratio_c_variables", num_continuous / (cur_numcols+0.0));
	writefeat("ratio_s_variables", num_semicont / (cur_numcols+0.0));
	writefeat("ratio_n_variables", num_semiint / (cur_numcols+0.0));
	
	writefeat("num_i+_variables", (*variable_sets)[0].size());
	writefeat("ratio_i+_variables", ((*variable_sets)[0].size() + 0.0) / (cur_numcols+0.0));
	
	/* Compute support sizes for integer variables and percentage of unbounded integer variables */
	double *support_size = new double[(*variable_sets)[0].size()];
	if (probtype == CPXPROB_LP || probtype == CPXPROB_QP || probtype == CPXPROB_QCP){
		/* These problem types have only continuous variables, so these features are not active (they are conditional!)*/
		num_unbounded_discrete = 0;
	} else {
		/* Percentage continuous or semi-continuous variables, and support sizes for non-continuous  variables. */
		num_non_cont = 0;
		num_unbounded_discrete = 0;
		set<int>::iterator it;

		for (it=(*variable_sets)[0].begin(); it!=(*variable_sets)[0].end(); it++){
			i = *it;
			if (xctype[i] == 'B'){
				support_size[num_non_cont++] = 1+1;
			} else {
				if (xctype[i] == 'I' || xctype[i] == 'N'){
					if (up_bound[i] == CPX_INFBOUND || low_bound[i] == -CPX_INFBOUND){
						num_unbounded_discrete++;
					} 
				}
				
				if (xctype[i] == 'I'){
					support_size[num_non_cont++] = up_bound[i]-low_bound[i]+1;
				} else {
					if (xctype[i] == 'N'){ /* semi-integer */
						support_size[num_non_cont++] = up_bound[i]-low_bound[i]+1+1;
					} else {
						if (xctype[i] == 'S'){ 	/* semi-continuous: either 0 or in that interval, so "2" is the best support size I can come up with */
							support_size[num_non_cont++] = 2;
						} else {
							fprintf (stderr, "WARNING: unknown xctype at position %d: %c. Ignoring this variable in feature computation.\n", i, xctype[i]);
						}
					}
				}	
			}
		}
	}
	writefeat("num_unbounded_disc", num_unbounded_discrete);
	if ((*variable_sets)[0].size() == 0){
		writefeat("ratio_unbounded_disc", 0);
	} else {
		writefeat("ratio_unbounded_disc", (num_unbounded_discrete+0.0)/((*variable_sets)[0].size()));
	}

	/* Compute the stats for the support size vector. */
	compute_array_stats(support_size, num_non_cont, "support_size", tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}
}


void computeVcgFeatures(CPXENVptr env, CPXLPptr lp, set<int> variable_set, int var_set_num){
	int i,j,k_all,k;
	set<int>::iterator it;
	string var_set_numstr = convertInt(var_set_num);

	double *A_ij_normalized = new double[nzcnt];
	double *a_normalized_varcoefs = new double[cur_numrows];
	double *vcg_constraint_degree = new double[cur_numrows];
	double *vcg_constraint_sum    = new double[cur_numrows];

	double *obj_fun_coef_this = new double[variable_set.size()]; 
	/* The following are really integers, but to avoid having to rewrite computation of statistics for ints, we make them doubles. */
	double *vcg_var_degree = new double[variable_set.size()];
	double *vcg_var_sum    = new double[variable_set.size()];
	double *normed_abs_row_vals = new double[variable_set.size()];

	int *map_all_vars_to_active_vars = new int[cur_numcols];

	/* Compute map from all variables to active ones. VERY IMPORTANT! */
	for (k_all=0; k_all<cur_numcols; k_all++){
		map_all_vars_to_active_vars[k_all] = -1;
	}
	int num_active=0;
	for (it=variable_set.begin(); it!=variable_set.end(); it++){
		k_all = *it;
		map_all_vars_to_active_vars[k_all] = num_active++;
	}
	
	/****************************************************************************
	* Compute variable constraint graph features.
	****************************************************************************/
	/* Initialize counters. */
	for (i=0; i<cur_numrows; i++) vcg_constraint_degree[i] = 0;
	for (i=0; i<cur_numrows; i++) vcg_constraint_sum[i] = 0;
	for (k=0; k<variable_set.size(); k++) vcg_var_degree[k] = 0;
	for (k=0; k<variable_set.size(); k++) vcg_var_sum[k] = 0;

	/* Enumerate nonzero entries, get variable and constraint degrees, and compute features of A matrix */
	int count = 0;
	for(i=0; i<cur_numrows; i++){
		int for_up_bound;
		if (i==cur_numrows-1) {
			for_up_bound = nzcnt; 	/* last row, special case for the upper limit of the loop */
		} else {
			for_up_bound = rmatbeg[i+1];
		}
		double sum_abs = 0;
		int num_nz_this_row = 0; // only consider columns in the current variable set.
		for (j=rmatbeg[i]; j<for_up_bound; j++){
			k_all = rmatind[j];
			/* If variable k_all is not an element of the set we're currently considering then skip it. */
			k = map_all_vars_to_active_vars[k_all];
			if (k == -1){
				continue;
			}
							
			vcg_constraint_degree[i]++;
			vcg_var_degree[k]++; 

			/* For computing features of the rows of the A matrix */
			double a_ik_all = rmatval[j];
			normed_abs_row_vals[num_nz_this_row++] = fabs(a_ik_all); // absolute of nonzero entries in this row (only considering variables in the current set).
			sum_abs += fabs(a_ik_all);
			vcg_var_sum[k] += a_ik_all;
			vcg_constraint_sum[i] += a_ik_all;

			if (rhs_b[i] > 1e-6 || rhs_b[i] < -1e-6){
				/* A_ij_normalized holds constraint coefficients of normalized MIP with b = vector of ones (ignoring those rows with b=0 */
				A_ij_normalized[count++] = a_ik_all / rhs_b[i];
			}

			//printf(" row=%d, j=%d, k=%d, a_ik=%lf, rhs=%lf, begin=%d, end=%d,#vars in constraint=%d\n", i, j, k, a_ik, rhs_b[i], rmatbeg[i], for_up_bound , (for_up_bound - rmatbeg[i] - 1));
		}
		
		/* Compute variation coefficient of normalized absolute nonzero entries in this row */
		for (j=0; j<num_nz_this_row; j++){
			normed_abs_row_vals[j] /= sum_abs;
		}
		if (num_nz_this_row == 0){
			a_normalized_varcoefs[i] = 0;
		} else {
			a_normalized_varcoefs[i] = d_var_coef(normed_abs_row_vals, num_nz_this_row);
		}	
	}
	
	/* Compute the stats for VCG_constr_deg. */
	compute_array_stats(vcg_constraint_degree, cur_numrows, "vcg_constr_deg" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}
	
	/* Compute the stats for VCG_var_deg. */
	compute_array_stats(vcg_var_degree, variable_set.size(), "vcg_var_deg" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}
	
	/* Compute the stats for VCG_constr_weight. */
	compute_array_stats(vcg_constraint_sum, cur_numrows, "vcg_constr_weight" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}

	/* Compute the stats for VCG_var_weight. */
	compute_array_stats(vcg_var_sum, variable_set.size(), "vcg_var_weight" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}

	/* Compute the stats for A_ij_normalized. */
	compute_array_stats(A_ij_normalized, count, "A_ij_normalized" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}
	
	/* Compute the stats for A_normalized_varcoefs. */
	compute_array_stats(a_normalized_varcoefs, cur_numrows, "A_norm_varcoefs" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}

	
	/****************************************************************************
	* Compute features using objective function coefficients. 
	****************************************************************************/
	int num_constrained_vars = 0;
	double* mean_obj_coef_per_constr = new double[variable_set.size()];
	double* mean_obj_coef_per_sqrt_constr = new double[variable_set.size()];
	for (it=variable_set.begin(); it!=variable_set.end(); it++){
		k_all = *it;
		k = map_all_vars_to_active_vars[k_all];
		obj_fun_coef_this[k] = obj_fun_coef[k_all];
		if (vcg_var_degree[k] == 0){
			continue;
		} else {
			mean_obj_coef_per_constr[num_constrained_vars] = obj_fun_coef[k] / vcg_var_degree[k];
			mean_obj_coef_per_sqrt_constr[num_constrained_vars++] = obj_fun_coef[k] / sqrt(vcg_var_degree[k]);
		}
	}

	/* Compute the stats for objective function vector. */
	compute_array_stats(obj_fun_coef_this, variable_set.size(), "obj_coefs" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}
	
	// For these 2 vectors: original version only used stddev.
	/* Compute the stats. Origin of this feature: WDP, price divided by number of bids. */
	compute_array_stats(mean_obj_coef_per_constr, num_constrained_vars, "obj_coefs" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}

	/* Compute the stats. Origin of this feature: WDP, price divided by sqrt(number of bids). */
	compute_array_stats(mean_obj_coef_per_sqrt_constr, num_constrained_vars, "obj_coefs" + var_set_numstr, tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		writefeat(tmp_stats_names[i], tmp_stats[i]);
	}
}



void computeVgFeatures(CPXENVptr env, CPXLPptr lp, set<int> variable_set, int var_set_num){
	// TODO.
	int i,j,k;
	string var_set_numstr = convertInt(var_set_num);

	double *vg_degree = new double[variable_set.size()];
	double *vg_degree_copy = new double[variable_set.size()];
	for (k=0; k<variable_set.size(); k++) vg_degree[k] = 0;
	vector< set<int> > vset = vector< set<int> > (variable_set.size());
	int **vg_neighbours = new int*[variable_set.size()];
}



/* probing features based on log file from CPLEX */
void prob_cplex(CPXENVptr env, CPXLPptr lp){
 	int status=0;
        CPXFILEptr fpout  = NULL;
	CPXCHANNELptr cpxresults, cpxwarning, cpxerror, cpxlog;
        char* probfile;
        probfile=new char[256];
  //      char* probfile2;
  //      probfile2=new char[256];
        double objval=hugeValue;
	double finalgap=hugeValue;
        double mip_obj=hugeValue;
	int itcnt ;

     	status = CPXsetdblparam (env, CPX_PARAM_TILIM, 5);
//		status = CPXsetintparam (env, CPX_PARAM_ITLIM, 3000);
	
      	sprintf(probfile, "%s", P_tmpdir);
        strcat(probfile, "/outputXXXXXX");
        mkstemp(probfile);
        fpout = CPXfopen (probfile, "w");
			
	//status = CPXgetchannels (env, &cpxresults, &cpxwarning, &cpxerror, &cpxlog);
	status = CPXsetlogfile (env, fpout);
	status = CPXmipopt (env, lp);

	itcnt = CPXgetmipitcnt (env, lp);
        // printf("\nitcnt =%d\n",  itcnt);
      	status = CPXgetmipobjval (env, lp, &mip_obj);
                       
	if ( status ) {
		if (status == 1217 ){
			printf("\nmip_obj does not exist\n");
		} else {
			fprintf (stderr, "CPXgetmipobjval failed, error code %d.\n", status);
			terminate(status);
		}
		} 
	else {
		// printf("\nmip_obj = %lf\n", mip_obj);
	}

	status = CPXgetbestobjval (env, lp, &objval);
	if ( status ) {
		fprintf (stderr, "CPXgetbestobjval failed, error code %d.\n", status);
		terminate(status);
	}
	else {
		// printf("\nobjval = %lf\n", objval);
	}
                        
                        
         if (mip_obj<hugeValue/2)
                  finalgap= abs(objval - mip_obj) / (1e-10 + abs(objval));
         // printf("\nfinalgap = %lf\n", finalgap);
       

	status = CPXflushstdchannels (env);
 	CPXfclose (fpout);
        // printf ("Finished writing to file.\n");

	processfile(probfile);
        remove(probfile);

         if (finalgap >1.0)
           writefeat("finalgap", 1.0);
         else
           writefeat("finalgap", finalgap);
                        

}	
 
void processfile(char* filename){
   // write stuff into file lin
        ifstream fin; 
        string mystring;
        string targetstr;
         
         
        fin.open(filename);
        vector<string> pretimeVec;
 	vector<double> pretimeDV;
        vector<string> rowVec;
        vector<string> colVec;
 	vector<string> nonzeroVec;
        vector<string> cliqueMemberVec;
        vector<string> cliqueCutVec;
        vector<string> iterVec;
        vector<string> dualObjVec;
        vector<string> reltimeVec;
        int pos;
        int col1;
        int col2;
        int col3;
        int col4;
        int col5;
        int col6; 
        int col0;
        int col7;
        int col8;
        vector<string> insolutionVec;
	vector<double> insolutionDV;
        vector<string> nodeVec;
        vector<string> nodenewVec;
        vector<string> nodeleftVec;
        vector<double> nodeDV;
        vector<double> nodenewDV;
        vector<double> nodeleftDV;
        vector<string> objVec;
        vector<string> iinfVec;
	vector<double> iinfDV;
        vector<string> bestinVec;
	vector<double> bestinDV;        
	vector<string> bestnodeVec;
        vector<double> bestnodeDV;
        vector<string> itcntVec;
        vector<double> itcntDV;
        vector<string> gapVec;
	vector<double> gapDV;
        vector<string> cutsVec;
	vector<double> cutsDV;
        vector<string> numcutsVec;
        vector<double> numcutsDV;
        vector<string> integralVec;
        vector<double> integralDV;
        vector<string> nodeobjVec;
        vector<double> nodeobjDV;
                
    	while (! fin.eof() ){
    			getline(fin, mystring);
			cout << "The string is " << mystring << endl;
       			targetstr="Presolve time =";
       			getinfo(mystring, targetstr, pretimeVec);
                        string targetstr="Reduced MIP has ";
			getinfo(mystring, targetstr, rowVec);
     			targetstr="rows, ";
			getinfo(mystring, targetstr, colVec);
     			targetstr="columns, and ";
			getinfo(mystring, targetstr, nonzeroVec);
     			targetstr="Clique table members:";
			getinfo(mystring, targetstr, cliqueMemberVec);
			targetstr="Clique cuts applied:";
			getinfo(mystring, targetstr, cliqueCutVec);
     			targetstr="Iteration: ";
			getinfo(mystring, targetstr, iterVec);
			targetstr="Dual objective     =   ";
			getinfo(mystring, targetstr, dualObjVec);
			targetstr="Root relaxation solution time =  ";
			getinfo(mystring, targetstr, reltimeVec);
                        //now, we need to deal with the table. (note: they may have two or more of tables)
                        targetstr="   Node ";
                        pos = mystring.find(targetstr);
                        if (pos>=0){
                      //      cout << "find the table \n";
                            col0=0;
                            targetstr="   Node";
                            col1=mystring.find(targetstr)+targetstr.size();
                            targetstr=" Left";
                            col2=mystring.find(targetstr)+targetstr.size();   
                            targetstr=" Objective";
                            col3=mystring.find(targetstr)+targetstr.size();  
                            targetstr=" IInf";
                            col4=mystring.find(targetstr)+targetstr.size();  
                            targetstr=" Best Integer";
                            col5=mystring.find(targetstr)+targetstr.size();  
                            targetstr=" Best Node";
                            col6=mystring.find(targetstr)+targetstr.size();  
			    targetstr=" ItCnt";
                            col7=mystring.find(targetstr)+targetstr.size();  
                            targetstr=" Gap";
                            col8=mystring.find(targetstr)+targetstr.size();    
                            getline(fin, mystring);
			    cout << "The string is " << mystring << endl;
                            int take=1;
                            while (take>0){
                              getline(fin, mystring);
                              cout << "The string is " << mystring << endl;
                              pos=mystring.find("                                                                ");
                              
                              if (pos>=0 || mystring.empty()){
                                  take=0;
                                  break;
                              }
                              pos=mystring.find("Elapsed");
                              if (pos<0){
				getinfosubstr(mystring, 0, 1,  insolutionVec);
				getinfosubstr(mystring, 1, col1-1,  nodeVec);
				getinfosubstr(mystring, col1, 1,  nodenewVec);
				getinfosubstr(mystring, col1+1, col2-col1,  nodeleftVec);
				getinfosubstr(mystring, col2+1, col3-col2,  objVec);
				getinfosubstr(mystring, col3+1, col4-col3,  iinfVec);
				getinfosubstr(mystring, col4+1, col5-col4,  bestinVec);
				getinfosubstr(mystring, col5+1, col6-col5,  bestnodeVec);
				getinfosubstr(mystring, col6+1, col7-col6,  itcntVec);
				getinfosubstr(mystring, col7+1, col8-col7-1,  gapVec);
			      }                       
                            }
                        }
                        
         }
         fin.close();
       
        int type1 =1;
        int type2=2;
        
        double bigvalue=1e100;
        double smallvalue = 0.0;
        double noone=1234.1234;
        double mymean, mymax, mymin, mystd, mysum, counter, mytmp;
        double maxit;

        

        str2double(itcntVec, itcntDV, type1, smallvalue, smallvalue);
        basicVect(itcntDV, noone, mysum, mymean, mymax, mymin, mystd);
        maxit = mymax;
        writefeat("numit",maxit);

        str2double(insolutionVec, insolutionDV, type2, smallvalue, smallvalue);
        basicVect(insolutionDV, noone, mysum, mymean, mymax, mymin, mystd);
        writefeat("numinsolution",mysum);
  
        str2double(nodeVec, nodeDV, type1, smallvalue, smallvalue);
        basicVect(nodeDV, noone, mysum, mymean, mymax, mymin, mystd);
        writefeat("nodeprob",mymax);
  
        str2double(nodenewVec, nodenewDV, type2, smallvalue, smallvalue);
        basicVect(nodenewDV, noone, mysum, mymean, mymax, mymin, mystd);
        writefeat("newin",mysum);
  
        str2double(nodeleftVec, nodeleftDV, type1, smallvalue, smallvalue);
        basicVect(nodeleftDV, smallvalue, mysum, mymean, mymax, mymin, mystd);
        writefeat("nodeleft_max",mymax);
 	writefeat("nodeleft_min",mymin);
        writefeat("nodeleft_mean",mymean);
        writefeat("nodeleft_std",mystd);
      
        splitoneVect(objVec, nodeobjVec, integralVec);
        str2double(nodeobjVec, nodeobjDV, type1, hugeValue, hugeValue);
        basicVect(nodeobjDV, hugeValue, mysum, mymean, mymax, mymin, mystd);
        writefeat("obj_mean", mymean);
        writefeat("obj_max", mymax);
        writefeat("obj_min", mymin);
        writefeat("obj_std", mystd);

        str2double(integralVec, integralDV, type2, smallvalue, smallvalue);
        countitemVect(integralDV, 1.0, counter);
        writefeat("numfeas", counter);

	str2double(iinfVec, iinfDV, type2, smallvalue, smallvalue);
        basicVect(iinfDV, smallvalue, mysum, mymean, mymax, mymin, mystd);
        writefeat("iinf_max",mymax);
 	writefeat("iinf_min",mymin);
        writefeat("iinf_mean",mymean);
        writefeat("iinf_std",mystd);
      
	str2double(bestinVec, bestinDV, type1, hugeValue, hugeValue);
        basicVect(bestinDV, hugeValue, mysum, mymean, mymax, mymin, mystd);
        writefeat("bestin_max",mymax);
 	writefeat("bestin_min",mymin);
        writefeat("bestin_mean",mymean);
        writefeat("bestin_std",mystd);
      
        splitVect(bestnodeVec, cutsVec, numcutsVec);
        str2double(bestnodeVec, bestnodeDV, 1, hugeValue, hugeValue);
        basicVect(bestnodeDV, hugeValue, mysum, mymean, mymax, mymin, mystd);
        writefeat("nodeobj_max", mymax);
        writefeat("nodeobj_min", mymin);
        writefeat("nodeobj_std", mystd);
        writefeat("nodeobj_avg", mymean);

        str2double(numcutsVec, numcutsDV, 1, hugeValue, hugeValue);
	basicVect(numcutsDV, hugeValue, mysum, mymean, mymax, mymin, mystd);
 	writefeat("numcuts_max", mysum);
  
        str2double(gapVec, gapDV, 1, hugeValue, hugeValue);
        basicVect(gapDV, hugeValue, mysum, mymean, mymax, mymin, mystd);
        writefeat("gap_max", mymax);
        writefeat("gap_min", mymin);
        writefeat("gap_std", mystd);
        writefeat("gap_avg", mymean);
        writefeat("gap_red", mymax-mymin);
        writefeat("pre_t", readfirst(pretimeVec));
        writefeat("rel_t", readfirst(reltimeVec));
	writefeat("new_row", readfirst(rowVec));
	writefeat("new_col", readfirst(colVec));
	writefeat("new_nonzero", readfirst(nonzeroVec));
	writefeat("clique_table", readfirst(cliqueMemberVec));

        
}



//vector basic information
void basicVect(vector<double> &dv, double notcount, double &mysum, double &mymean, double &mymax, double &mymin, double &mystd){
      double tmp;
      mymin=-512.0;
      mymax=-512.0;
      mysum=0.0;
      double mysqrsum=0.0;
      mymean=-512.0;
      mystd=-512.0;
      int index=0;
     
      vector<double>::iterator i;
     for (i = dv.begin(); i != dv.end(); i++) {
                tmp=*i;
                if (tmp!=notcount){
                    if (index==0){
                        mymax=tmp; 
		        mymin=tmp;
                    }
                       
                    mysum=mysum+tmp;
                    mysqrsum=mysqrsum+tmp*tmp;
                    index++;
                    if (tmp>=mymax)
                       mymax=tmp;
                    if (tmp <=mymin)
                        mymin=tmp;
                 }
     }
     if (index>0){
       mymean= mysum/index;
       mystd=sqrt((mysqrsum-index*mymean*mymean)/index);
   //pay attention to num. problem.
       mystd=max(mystd, 0.0001);    
   }
}

void countitemVect(vector<double> dv, double item, double & numitem){
      double tmp;
      numitem=0.0;
      vector<double>::iterator i;
      for (i = dv.begin(); i != dv.end(); i++) {
                tmp = *i;
                if (tmp==item){
		      numitem++;
           	}
     }

}



//split vector string for cuts/best node
void splitVect(vector<string> &sv, vector<string> &sv1, vector<string> &sv2){	
	vector<string>::iterator i;
        string s1, s2, tmp;
        int pos;
        
	for (i = sv.begin(); i != sv.end(); i++) {
                tmp=*i;
		if (tmp.find(":")==string::npos){
                    sv1.push_back("");
                    sv2.push_back("");
		   
		}
       		else {
		    pos=tmp.find(":");
		    sv1.push_back(tmp.substr(0, pos+1));
                    sv2.push_back(tmp.substr(pos+1));
                     *i="";
 //                   cout << sv1.back() << " " << sv2.back() << endl;
                }
	}
}

//split vector string for objection since it may contain integral etc.
void splitoneVect(vector<string> &sv, vector<string> &sv1, vector<string> &sv2){	
	vector<string>::iterator i;
        string tmp;
        
	for (i = sv.begin(); i != sv.end(); i++) {
                tmp=*i;
		if (tmp.find(".")==string::npos){
                    sv1.push_back("");
                    sv2.push_back(tmp);
		}
       		else {
		    sv1.push_back(tmp);
                    sv2.push_back("");
                }
	}
}

// transfer string to doulble with different refill option
// 1: refill with previous value
// 2: refill with bigvalue

void str2double(vector<string> &sv, vector<double> &dv, int refill, double firstline, double bigvalue){
vector<string>::iterator i;
double tmp;
double foo;
string stmp;
if (refill == 1){    
  for (i = sv.begin(); i != sv.end(); i++) {
    string stmp=*i;
    strip(stmp);
    if (stmp.empty()){
        
        if (i !=sv.begin()){
           foo=dv.back();
           dv.push_back(foo);}
        else
           dv.push_back(firstline);
        }
       else {
    tmp =  str_to_f(*i);
    dv.push_back(tmp);
      }
  }
 }
if (refill ==2){
 for (i = sv.begin(); i != sv.end(); i++) {
    stmp=*i;
    strip(stmp);
   
    if (stmp.empty()){
        
        if (i !=sv.begin()){
          
           dv.push_back(bigvalue);}
        else
           dv.push_back(firstline);
        }
       else {
    tmp =  str_to_f(*i);
    dv.push_back(tmp);
      }
  }
 }
if (refill >2)
 cout << "Error: Refill must be 1 or 2! \n";
 
}

void getinfosubstr(string mystring, int ss, int ee,  vector<string> &sv){
	string tmpstr;
        string newstr;
        newstr=mystring.substr(ss, ee);
        sv.push_back(newstr);
}

void getinfo(string mystring, string targetstr, vector<string> &sv){
	int pos1;
        pos1 = mystring.find (targetstr);
        string newstr;
        string tmpstr;
        if (pos1>=0) {
		newstr=mystring.substr(pos1+targetstr.size());
 		stringstream ss(newstr);
                ss >> tmpstr;
                // strip(tmpstr);
                sv.push_back(tmpstr);
          }
}

void display(vector<double> &sv){
  vector<double>::iterator i;
  for (i = sv.begin(); i != sv.end(); i++) {
    cout << *i << endl;
  }
  cout << "=================================" << endl;
}

void strip(string &s) {
  string::iterator i = s.begin();
  while (isspace(*i)) {
     s.erase(i);
     i = s.begin();
   }
  string::reverse_iterator j = s.rbegin();
  while (isspace(*i)) {
    s.resize(s.length()-1);
    j = s.rbegin();
  }
}


double readfirst(vector<string> dv){
      string tmp;
      double myvalue=-512.0;
      vector<string>::iterator i;
      for (i = dv.begin(); i != dv.end(); i++) {
                tmp = *i;
                myvalue=str_to_f(tmp);
                return myvalue;
     }
     return myvalue;
}

// transfer a string to double
double str_to_f(string newstr){
 double myvalue ; 
 string tmp;
 istringstream in(newstr);
 double one=1.0;
 double two=2.0;
 double three=3.0;
 double four=4.0;
 double unknown=9.0;
// in >> tmp;

 if (newstr.find("*")!=string::npos){
    return one;
}
 if (newstr.find("+")!=string::npos)
    return one;


if (newstr.find("Cuts")!=string::npos)
    return one;
//if (newstr.find("Impl")!=string::npos)
//    return two;

 if (newstr.find("Cliques")!=string::npos)
    return two;

 if (newstr.find("integral")!=string::npos)
    return one;
 if (newstr.find(":")!=string::npos)
    return unknown;
  

 in >> myvalue;
 return myvalue;
}


// Now it is time for main

int main (int argc, char *argv[]){
	/* Declare variables. */
	CPXENVptr     env = NULL;
	CPXLPptr      lp = NULL;
	int probtype;

	vector< set<int> > variable_sets;
	vector< set<int> > constraint_sets;

	int           status = 0;
	int           i,j,k;
	int nzcnt_returned, surplus;
	
	double *all_clust_coef, clust_coef, deviation, num_neighbour_edges;
	double prices_stddev, price_per_number_stddev, price_per_sqrt_number_stddev, *tmp_double_cur_numcols;
	double time_clust_coef, time_relax, time_total;
	bool compute_vg = true;
	bool compute_clust_coef = true;
	bool compute_relax = true;
	bool debug = false;

	sw.Start();

	/* Check the command line arguments */
	if ( argc != 2 ) {
		usage (argv[0]);
		terminate(status);
	}

	/* Initialize the problem. */
	initialize(argv[1], debug, &env, &lp);
	
	/* Get problem type. */
	probtype = CPXgetprobtype (env, lp);
	writefeat("probtype", probtype);

	/* FH TODO: we're outputting basic information here. This information
           should instead be computed and outputted for all variable subsets. */

	/* Ask CPLEX for the problem size.  cur_numrows and cur_numcols store the 
	    current number of rows and columns, respectively.  */
	cur_numcols = CPXgetnumcols (env, lp);
	cur_numrows = CPXgetnumrows (env, lp);
	nzcnt = CPXgetnumnz(env, lp);   /* this does *not* include the nonzero entries for quadratic constraints */
	writefeat("n_vars", cur_numcols);
	writefeat("n_constr", cur_numrows);
	writefeat("n_nzcnt", nzcnt);

	/* Compute features for quadratically-constrained problems */
	int num_quad = CPXgetnumquad (env, lp); // number of variables that have quadratic objective coefficients
	int num_q_constr = CPXgetnumqconstrs (env, lp);
	int num_qpnz = CPXgetnumqpnz (env, lp);
	writefeat("nq_vars", num_quad);
	writefeat("nq_constr", num_q_constr);
	writefeat("nq_nzcnt", num_qpnz);

	/****************************************************************************
	* Get information about A and b.
	****************************************************************************/
	/* Get objective function coefficients. */
	obj_fun_coef = new double[cur_numcols];
	status = CPXgetobj (env, lp, obj_fun_coef, 0, cur_numcols-1);
	if ( status ) {
		fprintf (stderr, "Failed to get objective function coefficients. Error %d\n", status);
		terminate(status);
	}
	
	/* Get right hand side, the b array. */
	rhs_b = new double[cur_numrows];
	status = CPXgetrhs (env, lp, rhs_b, 0, cur_numrows-1);
	if ( status ) {
		fprintf (stderr, "Failed to get right hand size, the b array. Error %d\n", status);
		terminate(status);
	}

	/* Get rmatbed, rmatind, and rmatval. */
	rmatbeg = new int[cur_numrows];
	rmatind = new int[nzcnt];
	rmatval = new double[nzcnt];
	status = CPXgetrows(env, lp, &nzcnt_returned, rmatbeg, rmatind, rmatval, nzcnt, &surplus, 0, cur_numrows-1);
	if( nzcnt_returned != nzcnt ){
		printf("nzcnt = %d, nzcnt_returned = %d\n", nzcnt, nzcnt_returned);
		fprintf(stderr, "Error in computation of nonzero entries of the constraint matrix -- value returned by CPXgetrows different than the one from CPXgetnumnz\n" );
		terminate(status);
	}
	if( surplus < 0 ){
		fprintf(stderr, "negative surplus %d, array too small\n", surplus);
		terminate(status);
	}	
	
	/* Get variable types. */
	getVariableTypes(env, lp, probtype, cur_numcols, &variable_sets);
	
	/* Get constraint sets (<=, =, >=, 'R' */
	getConstraintTypes(env, lp, cur_numrows, &constraint_sets);


	/****************************************************************************
	* Compute basic information for each constraint type.
	****************************************************************************/
	for (int const_set_num = 0; const_set_num<3; const_set_num++){
		string const_set_numstr = convertInt(const_set_num);

		/* Get RHS vector for this constraint type. */
		double *rhs_b_this = new double[constraint_sets[const_set_num].size()];
		int num_rhs_b_this = 0;
		for (set<int>::iterator it=constraint_sets[const_set_num].begin(); it!=constraint_sets[const_set_num].end(); it++){
			i = *it;
			rhs_b_this[num_rhs_b_this++] = rhs_b[i];
		}

		/* Compute the stats. */
		compute_array_stats(rhs_b_this, num_rhs_b_this, "rhs_c_" + const_set_numstr, tmp_stats, tmp_stats_names);
		for (i=0; i<num_stats_of_vector; i++){
			writefeat(tmp_stats_names[i], tmp_stats[i]);
		}
	}


	/****************************************************************************
	* Compute information for each variable type.
	****************************************************************************/
	for (int var_set_num = 0; var_set_num<3; var_set_num++){
		computeVcgFeatures(env, lp, variable_sets[var_set_num], var_set_num);
	}
	double time_VCG = sw.Lap();
	if (debug) fprintf(stderr, " time_VCG = %lfs\n", time_VCG);

	computeVgFeatures(env, lp, variable_sets[0], 0);
	double time_VG = sw.Lap();
	if (debug) fprintf(stderr, " time_VG = %lfs\n", time_VCG);

        prob_cplex(env, lp);       
 
	// TODO: VG, clust coef, edge density etc. 

	terminate(0);
}

/* This simple routine frees up the pointer *ptr, and sets *ptr to NULL */
static void
free_and_null (char **ptr)
{
   if ( *ptr != NULL ) {
      free (*ptr);
      *ptr = NULL;
   }
} /* END free_and_null */

static void
usage (char *progname)
{
   fprintf (stderr,"Usage: %s filename\n", progname);
   fprintf (stderr,"   where filename is a MIP instance file in .MPS, .LP, or .SAV format.\n");
   fprintf (stderr," Exiting...\n");
} /* END usage */
